'use strict';
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Document = exports.DEFAULT_VISIBLE_ELEMENTS = exports.Events = void 0;
const element_1 = require("./element");
const eventemitter3_1 = __importDefault(require("eventemitter3"));
const bson_1 = require("bson");
const object_generator_1 = __importDefault(require("./object-generator"));
const utils_1 = require("./utils");
const _1 = require(".");
/**
 * The event constant.
 */
exports.Events = {
    Cancel: 'Document::Cancel',
    Expanded: 'Document::Expanded',
    Collapsed: 'Document::Collapsed',
    VisibleElementsChanged: 'Document::VisibleElementsChanged',
    EditingStarted: 'Document::EditingStarted',
    EditingFinished: 'Document::EditingFinished',
    MarkedForDeletion: 'Document::MarkedForDeletion',
    DeletionFinished: 'Document::DeletionFinished',
};
/**
 * The id field.
 */
const ID = '_id';
exports.DEFAULT_VISIBLE_ELEMENTS = 25;
/**
 * Represents a document.
 */
class Document extends eventemitter3_1.default {
    /**
     * Send cancel event.
     */
    cancel() {
        // Cancel will remove elements from iterator, clone it before iterating
        // otherwise we will skip items
        for (const element of Array.from(this.elements)) {
            element.cancel();
        }
        this.emit(exports.Events.Cancel);
    }
    /**
     * Create the new document from the provided object.
     *
     * @param {Object} doc - The document.
     * @param {boolean} cloned - If it is a cloned document.
     */
    constructor(doc, cloned = false) {
        super();
        this.size = null;
        this.expanded = false;
        this.maxVisibleElementsCount = exports.DEFAULT_VISIBLE_ELEMENTS;
        this.editing = false;
        this.markedForDeletion = false;
        this.uuid = new bson_1.UUID().toHexString();
        this.doc = doc;
        this.cloned = cloned || false;
        this.isUpdatable = true;
        this.type = 'Document';
        this.currentType = 'Document';
        this.elements = this._generateElements();
    }
    apply(doc) {
        if (typeof doc?.generateObject === 'function') {
            doc = doc.generateObject();
        }
        const updatedKeys = [];
        let prevKey = null;
        for (const [key, value] of Object.entries(doc)) {
            if (this.get(key)) {
                this.get(key).edit(value);
            }
            else if (prevKey) {
                this.insertAfter(this.get(prevKey), key, value);
            }
            else {
                this.insertBeginning(key, value);
            }
            prevKey = key;
            updatedKeys.push(key);
        }
        for (const el of [...this.elements]) {
            if (!updatedKeys.includes(el.currentKey)) {
                el.remove();
            }
        }
    }
    /**
     * Generate the javascript object for this document.
     *
     * @returns {Object} The javascript object.
     */
    generateObject(options) {
        return object_generator_1.default.generate(this.elements, options);
    }
    /**
     * Generate the javascript object with the original elements in this document.
     *
     * @returns {Object} The original javascript object.
     */
    generateOriginalObject(options) {
        return object_generator_1.default.generateOriginal(this.elements, options);
    }
    /**
     * Generate the `query` and `updateDoc` to be used in an update operation
     * where the update only succeeds when the changed document's elements have
     * not been changed in the background.
     *
     * `query` and `updateDoc` may use $getField and $setField if field names
     * contain either `.` or start with `$`. These operators are only available
     * on MongoDB 5.0+. (Note that field names starting with `$` are also only
     * allowed in MongoDB 5.0+.)
     *
     * @param keyInclusionOptions Specify which fields to include in the
     *     originalFields list.
     *
     * @returns {Object} An object containing the `query` and `updateDoc` to be
     * used in an update operation.
     */
    generateUpdateUnlessChangedInBackgroundQuery(opts = {}) {
        // Build a query that will find the document to update only if it has the
        // values of elements that were changed with their original value.
        // This query won't find the document if an updated element's value isn't
        // the same value as it was when it was originally loaded.
        const originalFieldsThatWillBeUpdated = object_generator_1.default.getQueryForOriginalKeysAndValuesForSpecifiedFields(this, opts, true);
        const query = {
            _id: this.getId(),
            ...originalFieldsThatWillBeUpdated,
        };
        const updateDoc = object_generator_1.default.generateUpdateDoc(this);
        return {
            query,
            updateDoc,
        };
    }
    /**
     * Get an element by its key.
     *
     * @param {String} key
     *
     * @returns {Element} The element.
     */
    get(key) {
        return this.elements.get(key);
    }
    /**
     * Get an element by a series of segment names.
     *
     * @param {Array} path - The series of fieldnames. Cannot be empty.
     *
     * @returns {Element} The element.
     */
    getChild(path) {
        if (!path) {
            return undefined;
        }
        let element = this.elements.get(path[0]);
        let i = 1;
        while (i < path.length) {
            if (element === undefined) {
                return undefined;
            }
            element =
                element.currentType === 'Array'
                    ? element.at(path[i])
                    : element.get(path[i]);
            i++;
        }
        return element;
    }
    /**
     * Get the _id value for the document.
     *
     * @returns {Object} The id.
     */
    getId() {
        const element = this.get(ID);
        return element ? element.generateObject() : null;
    }
    /**
     * Generate the query javascript object reflecting the elements that
     * are specified by the keys listed in `keys`. The values of this object are
     * the original values, this can be used when querying for an update based
     * on multiple criteria.
     *
     * @param keyInclusionOptions Specify which fields to include in the
     *     originalFields list.
     *
     * @returns {Object} The javascript object.
     */
    getQueryForOriginalKeysAndValuesForSpecifiedKeys(opts = {}) {
        return object_generator_1.default.getQueryForOriginalKeysAndValuesForSpecifiedFields(this, opts, false);
    }
    /**
     * Get the _id value as a string. Required if _id is not always an ObjectId.
     *
     * @returns {String} The string id.
     */
    getStringId() {
        const element = this.get(ID);
        if (!element) {
            return null;
        }
        else if (element.currentType === 'Array' ||
            element.currentType === 'Object') {
            return JSON.stringify(element.generateObject());
        }
        return String(element.value);
    }
    /**
     * Insert a placeholder element at the end of the document.
     *
     * @returns {Element} The placeholder element.
     */
    insertPlaceholder() {
        return this.insertEnd('', '');
    }
    /**
     * Add a new element to this document.
     *
     * @param {String} key - The element key.
     * @param {Object} value - The value.
     *
     * @returns {Element} The new element.
     */
    insertBeginning(key, value) {
        const newElement = this.elements.insertBeginning(key, value);
        newElement._bubbleUp(element_1.Events.Added, newElement, this);
        this.emit(exports.Events.VisibleElementsChanged, this);
        return newElement;
    }
    /**
     * Add a new element to this document.
     *
     * @param {String} key - The element key.
     * @param {Object} value - The value.
     *
     * @returns {Element} The new element.
     */
    insertEnd(key, value) {
        const newElement = this.elements.insertEnd(key, value);
        newElement._bubbleUp(element_1.Events.Added, newElement, this);
        this.emit(exports.Events.VisibleElementsChanged, this);
        return newElement;
    }
    /**
     * Insert an element after the provided element.
     *
     * @param {Element} element - The element to insert after.
     * @param {String} key - The key.
     * @param {Object} value - The value.
     *
     * @returns {Element} The new element.
     */
    insertAfter(element, key, value) {
        const newElement = this.elements.insertAfter(element, key, value);
        newElement?._bubbleUp(element_1.Events.Added, newElement, this);
        this.emit(exports.Events.VisibleElementsChanged, this);
        return newElement;
    }
    /**
     * A document always exists, is never added.
     *
     * @returns Always false.
     */
    isAdded() {
        return false;
    }
    /**
     * Determine if the element is modified at all.
     *
     * @returns {Boolean} If the element is modified.
     */
    isModified() {
        for (const element of this.elements) {
            if (element.isModified()) {
                return true;
            }
        }
        return false;
    }
    /**
     * A document is never removed
     *
     * @returns {false} Always false.
     */
    isRemoved() {
        return false;
    }
    /**
     * The document object is always the root object.
     *
     * @returns {true} Always true.
     */
    isRoot() {
        return true;
    }
    /**
     * Generates a sequence of elements.
     *
     * @returns {Array} The elements.
     */
    _generateElements() {
        return new element_1.ElementList(this, this.doc);
    }
    /**
     * @deprecated Use DocumentEvents import instead
     */
    static get Events() {
        return exports.Events;
    }
    /**
     * Parse a new Document from extended JSON input.
     */
    static FromEJSON(input) {
        const parsed = bson_1.EJSON.parse(input, { relaxed: false });
        return new Document(parsed);
    }
    /**
     * Parse multiple Document from extended JSON input.
     * If the input consists of only a single document without
     * `[array]` brackets, return an array consisting of only
     * that document.
     */
    static FromEJSONArray(input) {
        const parsed = bson_1.EJSON.parse(input, { relaxed: false });
        return Array.isArray(parsed)
            ? parsed.map((doc) => new Document(doc))
            : [new Document(parsed)];
    }
    /**
     * Convert this Document instance into a human-readable EJSON string.
     */
    toEJSON(source = 'current', options = {}) {
        const obj = source === 'original'
            ? this.generateOriginalObject()
            : this.generateObject();
        return (0, utils_1.objectToIdiomaticEJSON)(obj, options);
    }
    /**
     * Expands a document by expanding all of its fields
     */
    expand() {
        this.expanded = true;
        for (const element of this.elements) {
            element.expand(true);
        }
        this.emit(exports.Events.Expanded);
        this.emit(exports.Events.VisibleElementsChanged, this);
    }
    /**
     * Collapses a document by collapsing all of its fields
     */
    collapse() {
        this.expanded = false;
        for (const element of this.elements) {
            element.collapse();
        }
        this.emit(exports.Events.Collapsed);
        this.emit(exports.Events.VisibleElementsChanged, this);
    }
    getVisibleElements() {
        return [...this.elements].slice(0, this.maxVisibleElementsCount);
    }
    setMaxVisibleElementsCount(newCount) {
        this.maxVisibleElementsCount = newCount;
        this.emit(exports.Events.VisibleElementsChanged, this);
    }
    getTotalVisibleElementsCount() {
        const visibleElements = this.getVisibleElements();
        return visibleElements.reduce((totalVisibleChildElements, element) => {
            return (totalVisibleChildElements + 1 + element.getTotalVisibleElementsCount());
        }, 0);
    }
    startEditing(elementId, field) {
        if (!this.editing) {
            this.editing = true;
            this.emit(_1.DocumentEvents.EditingStarted, elementId, field);
        }
    }
    finishEditing() {
        if (this.editing) {
            this.editing = false;
            this.emit(_1.DocumentEvents.EditingFinished);
        }
    }
    onUpdateStart() {
        this.emit('update-start');
    }
    onUpdateSuccess(doc) {
        this.emit('update-success', doc);
        this.finishEditing();
    }
    onUpdateBlocked() {
        this.emit('update-blocked');
    }
    onUpdateError(error) {
        this.emit('update-error', error.message);
    }
    markForDeletion() {
        if (!this.markedForDeletion) {
            this.markedForDeletion = true;
            this.emit(_1.DocumentEvents.MarkedForDeletion);
        }
    }
    finishDeletion() {
        if (this.markedForDeletion) {
            this.markedForDeletion = false;
            this.emit(_1.DocumentEvents.DeletionFinished);
        }
    }
    onRemoveStart() {
        this.emit('remove-start');
    }
    onRemoveSuccess() {
        this.emit('remove-success');
        this.finishDeletion();
    }
    onRemoveError(error) {
        this.emit('remove-error', error.message);
    }
}
exports.Document = Document;
exports.default = Document;
//# sourceMappingURL=document.js.map